package moca.core.proxy;

import java.io.IOException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.Collection;
import java.util.Iterator;

import moca.core.proxy.communication.CISEventProcessing;
import moca.core.proxy.communication.ClientCommunication;
import moca.core.proxy.communication.CommunicationProtocol;
import moca.core.proxy.communication.ServerCommunication;
import moca.core.proxy.communication.ServiceRegister;
import moca.core.proxy.development.ConnectionListener;
import moca.core.proxy.development.MessageListener;
import moca.core.proxy.development.ProxyFacade;
import moca.core.proxy.development.SubscribeListener;
import moca.core.proxy.message.Message;
import moca.core.proxy.state.Configurator;
import moca.core.proxy.util.ProxyConfiguration;
import moca.protocol.eci.server.ClientSubscription;

import org.apache.log4j.Logger;


/**
 * ProxyFramework is a singleton class responsible for starting servers for the communication 
 * with mobile clients. It configures the rules (context and actions), read from the XML
 * configuration file, that the proxy will be able to use to adapt or cache messages.
 * It also starts the communication with MOCA CIS (context Information Service)
 * 
 * <p>
 * Call @link #getInstance() to get access to this class.
 * After this, call  @link #init() to configure and launch the services provide by the proxy.
 * 
 * <p>
 *  To use proxy functionality you need the instance of ProxyFramework class.
 *
 * @author hana
 * 
 */

public class ProxyFramework implements ProxyFacade {
   /**
    * singleton instance
    * created when class is loaded.
    */
  	
	//static Category 			cat = Category.getInstance(ProxyFramework.class.getName());
   private static final Logger logger = Logger.getLogger(ProxyFramework.class);
   
   private static ProxyFramework instance = null;
   
	// Endereo local - proxy como cliente do servidor e do CIS
	String   clientLocalAddress; // endereo pblico local (para se comunicar com CIS e servidor da aplicao)
		
	/** Communication with clients */
	ClientCommunication	clientCommunic;
	
	/** Communication with server */
	ServerCommunication 	serverCommunic;
	
	/** Communication with CIS */
	CISEventProcessing	cisEventProcess = null;
	
	private String serverAddr = "127.0.0.1";
	private int serverPort = 60000;
	private InetSocketAddress 	serverSocket 	= null;
	
	/** Adaption proxy features */
	public Proxy proxy;
	
	
	/**
	 *   Private constructor, prevents direct instantiation of this class.
	 *  
	 */
	private ProxyFramework()  
	{
		clientLocalAddress = null;   //  endereo publico local...
		//proxy = Proxy.getInstance();
		clientCommunic = new ClientCommunication();
		serverCommunic = new ServerCommunication();
		
		
		configureProxyLocalAddress();		
	}

   /**
    * Get handle to the singleton
    * @return the singleton instance of this class.
    */
   public static ProxyFramework getInstance()
   {
	   if (instance == null)
		   instance = new ProxyFramework();
      return instance;
   }
	
	/**
	 * Try to get the proxy's local adresses configuration from the properties file.
	 * If this fails it tries to get the local address from local net, but this can
	 * be a problem, since there is no guarantee that the IP is public (it can be a private IP)  
	 */
	private void configureProxyLocalAddress(){
		
		try{
			String subscLocalAddr = ProxyConfiguration.getString("proxy.public_local.ip_address"); //$NON-NLS-1$
			clientLocalAddress = subscLocalAddr;
			serverAddr = ProxyConfiguration.getString("server.address");
			serverPort = Integer.parseInt(ProxyConfiguration.getString("server.port"));
			serverSocket = new InetSocketAddress(serverAddr,serverPort);
		}
		catch(Throwable e) {
         logger.error("Error reading proxy.properties file!",e);
         System.out.println("Error reading proxy.properties file." +
         		" \nTrying to get local address from localHost.\n");
			
         try {
				// Pega um endereo IP local 
         	// TODO -- explicar na documentao que isso pode dar errado pois
         	// ele pode pegar um IP privado, ou um IP que  barrado no firewall...
         	// por isso que deve configurar no proxy.properties.
					clientLocalAddress = 	InetAddress.getLocalHost().getHostAddress();
			} catch (UnknownHostException e1) {			
				logger.fatal("It was impossible to get an IP Address for the proxy",e1);
				System.exit(1);
			}
		}
		
		logger.info(" Proxy Address -->  as client at "	+ clientLocalAddress);
		
	}


	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#init(java.net.InetSocketAddress, java.net.InetSocketAddress)
	 */
	public void init(String proto, String mode) {
		
		// Creat elements for communication with the server
		
		if(mode.equals(CommunicationProtocol.SYNC)){
			if(proto.equals(CommunicationProtocol.TCP)){
				clientCommunic.startSynchronousTCPServer() ;
			     serverCommunic.SetSynchronousTCPSocket(serverSocket);
			}
			if(proto.equals(CommunicationProtocol.UDP)){
				clientCommunic.startSynchronousUDPServer();
				serverCommunic.SetSynchronousUDPSocket(serverSocket);
			}
		}
		
		if(mode.equals(CommunicationProtocol.ASYNC)){
			
			if(proto.equals(CommunicationProtocol.TCP)){
				clientCommunic.startAsynchronousTCPServer() ;
				serverCommunic.SetAsynchronousTCPSocket(serverSocket) ;
			}	
			
			else if(proto.equals(CommunicationProtocol.UDP)){
				clientCommunic.startSynchronousUDPServer();
				serverCommunic.SetSynchronousUDPSocket(serverSocket );
			}
		}
	
		if(mode.equals(CommunicationProtocol.EVENT)){
			try {
				 clientCommunic.startEventServer(proto) ;
				 serverCommunic.createSubscriberToServer(serverSocket , new InetSocketAddress(clientLocalAddress, 0), proto);
			} catch (IOException e1) {
				logger.error("Unable to create the subscriber to the server application. " + e1);
			}	
		}
		
		
		// Read from XML the configurations of the adatpations that will be 
		// inicialized
		Configurator conf = Configurator.getInstance();
	 
		//OBS: neste ponto as expresses j devem estar disponveis, pois vou precisar 
		// na hora de criar o objeto de comunicao com o CIS
		try {
	 		// Start subscriber for CIS
			cisEventProcess =  new CISEventProcessing(clientLocalAddress);
			cisEventProcess.configureExpressions(conf.getStateConfArray());
		} catch (IOException e) {
			logger.error("Impossible to subsbribe to CIS",e);
		}
		
		//Iniciar execuo do proxy
		proxy = Proxy.getInstance() ;
		proxy.run();
		
		//	Registrar no DS ??
		//registerProxyInMoCA(String serviceName)

	}

	
	/**
	 * Free all resources allocated by the proxy during its execution
	 * @throws IOException
	 */
	public void freeResources() {
		clientCommunic.freeResources() ;
		serverCommunic.freeResources() ;
		cisEventProcess.freeResources();
		proxy.unregisterAllClients();
		logger.debug("Resources freed..");
	}
	
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#addClient(java.lang.String, moca.core.proxy.communication.CommunicationProtocol, boolean)
	 */
	public void addClient(String newClient, CommunicationProtocol protocol,
			boolean registerAtCIS) {
		// Adiciona cliente a lista de clientes 
		addClient(newClient, protocol);
		// registra no CIS se desejado
		if(registerAtCIS)
			registerClientAtCIS(newClient);
	}

	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#addClient(java.lang.String, moca.core.proxy.communication.CommunicationProtocol)
	 */
	public void addClient(String newClient, CommunicationProtocol protocol) {
		logger.debug("Registering client  "+  newClient);
		ClientInfo clientInfo = new ClientInfo(newClient,protocol);
		proxy.registerClient(clientInfo);
	}
	
	public boolean isClientRegistered(InetSocketAddress address){
		String ret = proxy.getClientId(address) ;
		if(ret != null && ret.length() > 1)
			return true ;
		else
			return false ;
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#publishMsgToClients(moca.core.proxy.message.Message)
	 */
	public void publishMsgToClients(Message message) throws ProxyException {
		
		String subject = message.getSubject();
		 // The subject is set to uppercase before publishing the event.
	       subject = subject.toUpperCase();
	       
	       
		
		if(subject==null){
			logger.error("Impossible to publish the message for a null subject");
			throw new ProxyException("Invalid message subject - impossible to publish");
		}
		//Pegar lista de clientes interessados no assunto
		//Vector subscribedClients = (Vector) clientCommunic.getEventServer().getSubscribers(subject);
		Collection clientSubscriptions =  clientCommunic.getEventServer().getClientSubscriptionsCollection(subject) ;
			
		//Descobrir o identificador do cliente
		Iterator it = clientSubscriptions.iterator();
		while(it.hasNext()){		
			ClientSubscription aSubscription = (moca.protocol.eci.server.ClientSubscription) it.next();
			//InetSocketAddress clientAddress = aSubscriber.getSubscription().getSubscriberAddress(); verso anterior
			InetSocketAddress clientAddress = aSubscription.getSubscribedClient().getClientAddress() ;
			String clientId = proxy.getClientId(clientAddress);
				
			if (clientId != null){
				//Criar novas mensagens usando clone
				message.setAddressee(clientId);
				try {
					// Incluir na fila de msg a serem adaptadas
					proxy.addMessage(message.getAddressee(), (Message) message.clone());
				} catch (CloneNotSupportedException e) {
					logger.error("Impossible to create the message. "+ e);
				}
			}
			else // TODO else -- cliente no encontrado
				logger.warn("clientId == null. The sender must use the Message.setAddress(String) method");
		}

	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#publishMsgToServer()
	 */
	public void publishMsgToServer(Message message,CommunicationProtocol protocol) {
		try {
			this.serverCommunic.publishMessage(message, protocol.getProtocol());
		} catch (Exception e) {
			logger.error("Message " + message +" could not be published to the server. + " + e);	
		}
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#registerClientAtCIS(java.lang.String)
	 */
	public void registerClientAtCIS(String client) {
		this.cisEventProcess.subscribe(client);

	}

	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#registerProxyInMoCA(java.lang.String, java.util.Properties)
	 */
	public void registerProxyInMoCA(String serviceName) {
		// Registrar no DS
		
		//TODO -- Ver direito pois o DS no cadastra propriedades -- por enquanto
		// no fao nada se no se registyrar direito
		
		//	Registrar servio sncrono
		String serverIP = clientCommunic.getTcpServer().getLocalAddress().getHostName(); 
		int port = 	clientCommunic.getTcpServer().getLocalAddress().getPort();
		try {
			ServiceRegister.registerService(serverIP, "" + port, serviceName + "_SYNC");
		} catch (Exception e1) {
			logger.error("Impossible to register at DS." + e1);
		}
		//Registrar servio assncrono
		serverIP = clientCommunic.getEventServer().getLocalAddress().getHostName(); 
		port = 	clientCommunic.getEventServer().getLocalAddress().getPort();
		try {
			ServiceRegister.registerService(serverIP, "" + port, serviceName +"_ASYNC");
		} catch (Exception e2) {
			logger.error("Impossible to register at DS." + e2);
		}		
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#removeAllClients()
	 */
	public void removeAllClients() {
		//	Remover toda a lista de clientes 
		proxy.unregisterAllClients();
		
		// Remover do CIS
		this.cisEventProcess.unsubscribeAll();
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#removeAllClientsFromCIS()
	 */
	public void removeAllClientsFromCIS() {
		this.cisEventProcess.unsubscribeAll();
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#removeClient(java.lang.String)
	 */
	public void removeClient(String clientId) {		
		//Remover da lista de clientes 
		proxy.unregisterClient(clientId);
		// Removendo do CIS
		this.cisEventProcess.unsubscribe(clientId);		
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#removeClientFromCIS(java.lang.String)
	 */
	public void removeClientFromCIS(String client) {
		this.cisEventProcess.unsubscribe(client);

	}

	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#sendMsgToClient(moca.core.proxy.message.Message)
	 */
	public void sendMsgToClient(Message message)throws ProxyException {
		
		logger.debug("Proxy receive a message to be sent to client " + message.getAddressee());
		String receiver = message.getAddressee();
		if(receiver==null){
			logger.error("Impossible to send the message for a null client");
			throw new ProxyException("Invalid message addressee - impossible to send the message.");
		} 
		
		//Incluir a insero na fila de mensagens para adaptao!!!!!
		proxy.addMessage(message.getAddressee(), message);
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#sendMsgToServer(moca.core.proxy.message.Message)
	 */
	
	public void sendAsyncMsgToServer(Message m) throws IOException{
		serverCommunic.sendAsyncMsg(m);
	}
	public void sendMsgToServer(Message m, CommunicationProtocol protocol) {
		
		try {
			if(protocol.getProtocol().equals(CommunicationProtocol.TCP))
			     serverCommunic.sendMsg(m);
			else if(protocol.getProtocol().equals(CommunicationProtocol.UDP))
				 serverCommunic.sendAsyncMsg(m);
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} 
		// talvez passar o protocolo

	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#subscribeToServer(java.lang.String)
	 */
	public void subscribeToServer(String subject) {
		try {
			this.serverCommunic.subscribe(subject);
		} catch (ProxyException e) {
			logger.error("Impossible to subscribe to server. " + e );
		}
	}
	
	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#unsubscribeFromServer(java.lang.String)
	 */
	public void unsubscribeFromServer(String subject) {
		this.serverCommunic.unsubscribe(subject);

	}

	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#addServerMsgListener(moca.core.proxy.development.MessageListener)
	 */
	public void addServerMsgListener(MessageListener listener) {
		this.serverCommunic.addMessageListener(listener);
		
	}

	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#addClientMsgListener(moca.core.proxy.development.MessageListener)
	 */
	public void addClientMsgListener(MessageListener listener) {
		this.clientCommunic.addMessageListener(listener);
		
	}

	/* (non-Javadoc)
	 * @see moca.core.proxy.development.ProxyFacade#addClientConnectedListener(moca.core.proxy.development.ConnectionListener)
	 */
	public void addClientConnectedListener(ConnectionListener listener) {
		this.cisEventProcess.registerConnectionListener(listener);		
	}
	
	public void addClientSubscriptionListener(SubscribeListener subscListener){
		this.clientCommunic.addSubscribeListener(subscListener);
	}

	/**
	 * @return Returns the clientCommunic.
	 */
	public ClientCommunication getClientCommunication() {
		return clientCommunic;
	}
}