package moca.core.proxy.communication;

import java.io.IOException;
import java.net.InetSocketAddress;

import moca.core.proxy.ProxyException;
import moca.core.proxy.ProxyFramework;
import moca.protocol.net.TCPConnection;
import moca.protocol.net.async.tcp.AsyncTCPServer;

import org.apache.log4j.Logger;

/**
 * Indicate the communication mode and protocol of a client, and its net address.
 * Define the protocols accepted by the proxy, and the communication type. 
 * <p>
 * It must be used to register a client in the proxy system - @link ProxyFacade#addClient(String, CommunicationProtocol)
 *  
 * @author Hana Rubinsztejn
 *
 */
public class CommunicationProtocol {

	//static Category 	cat = Category.getInstance(CommunicationProtocol.class.getName());
	private static final Logger cat = Logger.getLogger(ProxyFramework.class);
	
   /** Constant that specifies a TCP protocol */
   public  final static String 	TCP = "TCP";
   /** Constant that specifies a UDP protocol */
   public  final static String 	UDP = "UDP";
   /** Constant that specifies a Event driven protocol */
   public  final static String 	EVENT = "EVENT";
   /** Constant that specifies a synchronous communication*/
   public final static String	SYNC = "SYNCHRONOUS";
   /** Constant that specifies an asynchronous communication*/
   public final static String	ASYNC = "ASYNCHRONOUS";

   /** The protocol used: TCP or UDP */
   private String protocol = null;
   
   /** The communication type: synchronous (SYNC) or asynchronous (ASYNC)*/
   private String communicationMode = null;
   
   /** The TCP Connection if it is a synchronous communication with TCP protocol */
   private TCPConnection tcpConn = null;

   /** the TCP Connection if it is a asynchronous communication. */
   private AsyncTCPServer connection = null ;
   
   /** Address - for UDP sync communication and asynchronous one*/
   private InetSocketAddress address = null; 
   
     
	/**
	 * @param protocol - CommunicationProtocol.TCP or CommunicationProtocol.UDP
	 * @param communicationMode - CommunicationProtocol.SYNC or CommunicationProtocol.ASYNC
	 * @throws ProxyException
	 */
	public CommunicationProtocol(String protocol, String communicMode) throws ProxyException {
				
		if((protocol == null || communicMode == null) || 
				!isValidCommunication(communicMode) || 
				!isValidProtocol(protocol))
			throw new ProxyException("Invalid communication protocol!!");
		
		this.protocol = protocol;
		this.communicationMode = communicMode;
	}
   
      
   
	/**
	 * @param protocol  The protocol to verify the validity
	 * @return The validity of the protocol
	 */
	public static boolean isValidProtocol(String protocol)
   {
		return (protocol.equalsIgnoreCase(TCP) || 
				  protocol.equalsIgnoreCase(UDP));
   }
	
	/**
	 * @param type The communication type to be verified
	 * @return  Returns if the commuication type is valid, i.e., 
	 *          if it is a synchronous or an asynchronous communication  
	 */ 
	public static boolean isValidCommunication(String type)
	{
		return type.equalsIgnoreCase(SYNC)|| type.equalsIgnoreCase(ASYNC) || type.equalsIgnoreCase(EVENT); 
	}
	
	
	/**
	 * @return Returns the tcpConnection.
	 */
	public TCPConnection getSyncTCPConnection() {
		return tcpConn;
	}
	
	public AsyncTCPServer getAsyncTCPConnection(){
		return connection ;
	}
	/**
	 * @param tcpConn The tcpConnection to set.
	 * @throws IOException
	 */
	public void setSyncTCPConnection(TCPConnection tcpConn)throws IOException {
		if ( protocol == TCP && communicationMode == SYNC){
			this.tcpConn = tcpConn;
		}
		else{
			cat.error("This cannot be used since its not a TCP synchronous communication");
			throw new IOException(
               "This method can not be used with the current communication protocol");
		}
	}	
	
	public void setAsyncTCPConnection(AsyncTCPServer tcpConn)throws IOException {
		if ( protocol == TCP && communicationMode == ASYNC){
			this.connection = tcpConn;
		}
		else{
			cat.error("This cannot be used since its not a TCP synchronous communication");
			throw new IOException(
               "This method can not be used with the current communication protocol");
		}
	}	
	/**
	 * The address - IP/Port
	 * @return Returns the address.
	 */
	public InetSocketAddress getAddress() {
		return address;
	}
	/**
	 * @param address The address to set.
	 */
	public void setAddress(InetSocketAddress address) {
		this.address = address;
	}
	
	
	/**
	 * @return Returns the communicationMode.
	 */
	public String getCommunicMode() {
		return communicationMode;
	}
	/**
	 * @param communicationMode The communicationMode to set.
	 */
	public void setCommunicMode(String communicMode) {
		this.communicationMode = communicMode;
	}
	/**
	 * @return Returns the protocol.
	 */
	public String getProtocol() {
		return protocol;
	}
	/**
	 * @param protocol The protocol to set.
	 */
	public void setProtocol(String protocol) {
		this.protocol = protocol;
	}
}
