
package moca.core.proxy.communication;

import moca.core.proxy.util.MocaProperties;
import moca.service.ds.client.ServiceAgent;

/**
 * This class register an service in the Discovery Service.
 * @author Hana
 */
public class ServiceRegister extends ServiceAgent
{
    // The instance of the object.
    private static ServiceRegister instance = null;
    

    // The service ip address.
    private String ip = "";
    
    // The service port.
    private int port = -1;
    
    // Constants list:
    public static String DEFAULT_SERVICE_TYPE = "MOCASERVICE";
    public static String DEFAULT_LANGUAGE = "ENGLISH";
    
    // The scope the service managed by this agent works at
    public static final String DEFAULT_SCOPE = "DEFAULT";
    
    // DS Address
    private static String dsAddress;
    
    // The internal constructor.
    private ServiceRegister() throws Exception
    {
        // It must be called.
        super();
        
        this.attrList    = "";
        this.curDAServer = "";
        this.language    = DEFAULT_LANGUAGE;
        this.scope       = DEFAULT_SCOPE;
        this.serviceType = DEFAULT_SERVICE_TYPE;
        
  		MocaProperties dsProps = new MocaProperties();
  		dsAddress = dsProps.getDsServerHost();
  		if (dsAddress == null)
  				throw new Exception("DS not configured...");
    }
    
    /**
     * Get the instance of ServiceRegister class.
     * @return ServiceRegister singleton.
     * @throws Exception
     */
    public static ServiceRegister getInstance() throws Exception
    {
        if(instance == null)
        {
         instance = new ServiceRegister();
        }
        return instance;
    }
    
    /**
     * Sets the IP address or the hostname of the
     * service.
     * @param ip
     */
    public void setIP(String ip)
    {
        if(ip == null)
            throw new IllegalArgumentException("The IP argument cannot be null");
        if(ip.equals(""))
            throw new IllegalArgumentException("The mepty IP or hostname is not valid argument.");
        this.ip = ip;
    }
    
    /**
     * Sets the port. The argument must be a valid port number.
     * 
     * @param port
     */
    public void setPort(String port)
    {    
        // validate the argument:
        if(port == null)
            throw new IllegalArgumentException("The port argument cannot be null.");
        if(port.equals(""))
            throw new IllegalArgumentException("The empty port number is invalid argument.");
        
        // convert the port from the string to int.
        int porta = Integer.parseInt(port);
        
        // verify invalid port number.
		if((porta < 0) && (porta > 65536))
			throw new IllegalArgumentException("The port number is invalid.");
		
		// set the port.
		this.port = porta;
    }
    
    /**
     * Sets the keyword. The keyword is used to find the service in the Discover
     * Service.
     * @param keyword is service attribute list.
     */
    public void setKeyword(String keyword)
    {
        // Verify if the argument is a valid argument:
        if(keyword == null)
            throw new IllegalArgumentException("Null keyword argument.");
        
        if(keyword.equals(""))
            throw new IllegalArgumentException("Empty keyword argument.");
        
        // Sets the keyword as attribute list.
        this.attrList = keyword;
    }

    /**
     * Sets the language used in the service.
     * @param language
     */
    public void setLanguage(String language)
    {
        // Verifies the argument validate.
        if(language == null)
            throw new IllegalArgumentException("Null language argument.");
        
        if(language.equals(""))
            throw new IllegalArgumentException("Empty language arment.");
        
        // Sets the language in the superclass property.
       this.language = language;
    }
    

    /**
     * Sets the Discovery Service address. This address will be used
     * to contact the Discovery Service.
     * @param dsAddress
     */
    public void setDSAddress(String dsAddress)
    {
        
        // Verifies the validates of the argument.
        if(dsAddress == null)
            throw new IllegalArgumentException("Null Discovery Service address argument.");
             
        // Sets the DS address in the local properties.
        this.curDAServer = dsAddress;
    }

    /**
     * Register the service in the Discovery Service. It is necessary to specify the IP address or the
     * hostname and port of the service. The service will be looked for by the keyword argument.
     * @param ip
     * @param port
     * @param keyword
     * @param dsAddress
     * @throws Exception
     */
    public static void registerService(String ip, String port, String keyword) throws Exception
    {
        registerService(ip, port, DEFAULT_LANGUAGE, keyword);
    }
    /**
     * Register the service in the Discovery Service. It is necessary to specify the IP address or the
     * hostname and port of the service. The service will be looked for by the keyword argument.
     * @param ip
     * @param port
     * @param keyword
     * @param dsAddress
     * @param language
     * @throws Exception
     */
    public static void registerService( String ip, String port, String language, String keyword) throws Exception
    {
        // Gets a reference to the ServiceRegister class. It will start
        // the properties with default values.
        ServiceRegister instance = ServiceRegister.getInstance();
        
        // Sets the ip.
        instance.setIP(ip);
        
        // Sets the port.
        instance.setPort(port);
        
        // Sets the language used in this service.
        instance.setLanguage(language);
        
        // Sets the keyword.
        instance.setKeyword(keyword);
        
        // Sets the Discovery Service Address.
        instance.setDSAddress(dsAddress);
        
        instance.setServiceType(DEFAULT_SERVICE_TYPE);
        
        // initialize the service to be registered.
        instance.initService();
        
    }
    
    /**
     * Set the service type.
     * @param string
     */
    private void setServiceType ( String servType )
    {
        if(servType == null)
            throw new IllegalArgumentException("The service type cannot be null.");
        
        if(servType.equals(""))
            throw new IllegalArgumentException("The service type cannot be empty.");
        
        this.serviceType = servType;
        
    }

    /**
     * Verify if the All service properties was set.
     */
    private void isAllServiceProperties() {
        
        if(this.attrList == null)
            throw new IllegalStateException("The service keyword was not set.");
        
        if(port == -1)
            throw new IllegalStateException("The service port was not set.");
        
        if(this.ip == null)
            throw new IllegalStateException("The service ip address or hostname was not set.");
        
        if(this.curDAServer == null)
            throw new IllegalStateException("The Discovery Service address was not set.");
    }

    /**
     * Initialize the service registration in the Discovery
     * Service.
     */
    public void initService()
    {
        // Verity if all service properties was set.
        this.isAllServiceProperties();
        
        // Creates the URL property.
        this.createURL();
        
        // Publish the service.
        instance.publService();
    }

    /**
     *  Creates the URL property with the service IP address
     * and the service port. The resultant string is "ip|port".
     * The ServiceFinder will use this format to contact the service.
     */
    private void createURL()
    {
        this.url = serviceType + "://"+this.ip+":"+this.port;        
    }

}
