/*
 * Created on 07/11/2004
 *
 * 
 */
package moca.core.proxy.communication.async;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.util.Iterator;
import java.util.Vector;

import moca.core.proxy.communication.ClientCommunication;
import moca.protocol.eci.server.ECIServer;
import moca.protocol.eci.server.SubscribedClient;
import moca.protocol.eci.server.ClientSubscription;
import moca.protocol.eci.server.dispatcher.EventDispatcherThread;
import moca.util.eci.DataMessage;
import moca.util.eci.EventProperties;

/**
 * This class extends @link moca.protocol.pubsub.eventprovider.EventProvider 
 * to add some functionalities needed by the proxy.
 * <p>
 * Publishs events to the subscribers which have been registered as interested in such events.
 * The EventProvider class facilitates the implementation of an event server. It will be used as
 * a event provider responsible to receive and processing all requests (Subscription, 
 * DataMessage and SubscriberRequest ) sent by the subscribers. 
 * <p>
 * The EventProvider implementation (aka event server) works as a forwarder when the subscriber calls one 
 * of the {@link moca.protocol.pubsub.subscriber.Subscriber#publish(String, Object)} methods.
 * The event server receives the <code>data</code> to be published and forward it to all
 * subscribers which has subscribed to the subject passed as argument to the {@link moca.protocol.pubsub.subscriber.Subscriber#publish(String, Object)}
 *  method. The data is encapsulated in a {@link moca.protocol.pubsub.DataMessage} object.  
 *<p>  
 * An event server can be implementing easily using this class. After creating a EventProvider object, the event server will
 * be able to publish data to the subscribers using the {@link #publish(String, Object)} or {@link #publish(String, EventProperty, Object)} methods.
 *
 * 
 * @author hana
 *
 */
public class MyEventProvider extends ECIServer {

	ClientCommunication clientCommunic = null;
	
	private Integer protocol = null ;
	/**
    * Cronstruct an event provider bound with specified local address. Every messages
    * will be received in this IP address and port.
    * <p>
	 * As the protocol has been omitted the event server will use both TCP and UDP as communication protocol.
	 *  That way, the publisher could process request from subscribers in the port specified using both protocols.
    * 
    * @param localAddress The local address to bind.
    * @param communic  The ClientCommunication that this provider is part of.
    * @throws IOException if the EventProvider could not be created
    */
		public MyEventProvider(InetSocketAddress localAddress, ClientCommunication communic, Integer protocol) throws IOException {
		super(protocol,localAddress);
		this.protocol = protocol ;
		clientCommunic = communic;
	}

	/**
	 * Cronstruct a publisher bound with specified local address. Every messages
	 * will be received in this IP address and port.
	 * <p>
	 * As the protocol has been omitted the publisher (or event server) will use both TCP and UDP as communication protocol.
	 *  That way, the publisher could process request from subscribers in the port specified using both protocols.
	 * <p>
	 * The second argument inform if the publisher will be able to publish data with event properties. 
	 * Enabling this feature to TRUE the publisher will implement a matching between the event propery object and
	 * the expression described in each subscription which has the same subject.  
	 * 
	 * @param localAddress The local address to bind
	 * @param isEXPRESSION_BASED if the publisher intend to publish data with event properties
	 * @param communic  The ClientCommunication that this provider is part of.
	 * @throws IOException if the EventProvider could not be created
	 */
	public MyEventProvider(InetSocketAddress localAddress, boolean isEXPRESSION_BASED, ClientCommunication communic, Integer protocol)
			throws IOException {
		super(protocol,localAddress, isEXPRESSION_BASED);
		this.protocol = protocol ;
		clientCommunic = communic;
	}

	
   // Metodos Novos
	
	// Esse mtodo  apenas para enviar o dado para um cliente especfico
   /**
    * Publish data to the subscriber if it has been subscribed as interested on the subject passed as argument.
    * The subscriber will receive the data encapsulated in an event {@link moca.protocol.pubsub.Event}. 
    * 
    * The data to be published must implement the Serializable interface 
    * 
    * @param subject subject to which the data will be published
    * @param data data to be published
    * @param subAddr Suscriber address (IP/Port) - indentifier of the client on the ECI 
    * @throws IOException if the data could not be published
    */
   public void publish(String subject, Object data,InetSocketAddress subAddr) 
   {
  		if(subject == null)
  		  throw new IllegalArgumentException("Invalid subject");

  		if(data == null)
  		    throw new IllegalArgumentException("Invalid data");

		this.publishData(subject,null,data,subAddr);    
   }
      
   /**
    * Publish data to the subscriber if it has been subscribed as interested on the subject passed as argument.
    * If the subscriber has specified a event selector expression, it will be notified if and only if the event properties
    * match the expression described in the subscription. The data published will be encapsulated in 
    * an event {@link moca.protocol.pubsub.Event}.
    * 
    * The data to be published must implement the Serializable interface 
    * 
    * @param subject subject to which the data will be published
    * @param data data to be published
    * @param eventProperty Event properties which will be used as base to evaluate the expressions of the subscribers
    * @param subAddr Suscriber address (IP/Port) - indentifier of the client on the ECI 
    * @throws IOException if the data could not be published
    */
   public void publish(String subject, EventProperties eventProperty, Object data,InetSocketAddress subAddr)
   {
 		if(subject == null)
 		   throw new IllegalArgumentException("Invalid subject");

  		if(data == null)
    		  throw new IllegalArgumentException("Invalid data");

  		if(eventProperty == null)
  		      throw new IllegalArgumentException("Invalid event property");

   		if(eventProperty.getAtributesHashTable().size() <= 0)
   		    throw new IllegalArgumentException("The event property is empty");
		
		publishData(subject,eventProperty,data, subAddr);
   }
   
   // Metodo novo, mas identico ao publishData da classe pai, exceto pela seleo das subscriptions
   // que devem ser atendidas -- no caso de um cliente especfico 
   
   protected void publishData(String subject, EventProperties eventProperty, Object data, InetSocketAddress clientAddr) 
   {
       Vector                   temp,clientSubscriptionsList = null;
       EventDispatcherThread    dispacherThread         = null;
       Thread                   thread                  = null;
   	Iterator  it = null;
       // ClientSubscription    currentClientSubscription = null;
       // Subscription          currentSubscription       = null;
       
       if(subject == null || data == null)
           return;
       
       // The subject is set to uppercase before publishing the event.
       subject = subject.toUpperCase();
       
       if(subscribedClientsList != null)
       {
    	   // Removes SubscribedClients that are no longer valid on the ECIServer due to a subscriptions confirmation timeout
           for(int i = 0; i < this.subscribedClientsList.size(); i++)
           {
               // Gets the current SubscribedClient from the subscribedClientsList
               SubscribedClient subscribedClient = (SubscribedClient)subscribedClientsList.get(i);
               
               if(!subscribedClient.isValidated())
               {
                   // Gets the current SubscrivedClient's ipAddress and port
                   String ipAddress = subscribedClient.getClientIpAddressStr();
                   String port      = subscribedClient.getClientPortStr();
                   
                   // Removes the SubscribedClient from the ECIServer with all its subscriptions
                   this.removeSubcribedClient(ipAddress,port);
                   
                   // The current element has been removed so lets decrement one step and continue to the next iteration
                   i--; continue;
               }
           }
       }
           
       // Gets the list of clientSubscriptions subscribed to the subject
       temp = (Vector) this.clientSubscriptionsHashTable.get(subject);
       
       clientSubscriptionsList = (Vector)temp.clone(); 
       
       if(clientSubscriptionsList == null)
           return; // Aborts the operation
       
       
   	it = clientSubscriptionsList.iterator();
   	while(it.hasNext()){		
   		ClientSubscription currentClientSubscription = (ClientSubscription)it.next();
   		SubscribedClient aSubscriber = currentClientSubscription.getSubscribedClient();
//   		SubscribedClient aSubscriber = (SubscribedClient) it.next();
   		if ( !(aSubscriber.getClientAddress().equals(clientAddr)) ){
   		//if ( !(aSubscriber.getClientIpAddressStr().equals(clientAddr)) ){
   			it.remove();
   		}
   	}				
       
       if(clientSubscriptionsList.isEmpty())
           return; // Aborts the operation
       
       // Creates and starts an EventDispatcherThread to publish events to the subscribers on the list
       dispacherThread = new EventDispatcherThread(subject, clientSubscriptionsList, data, eventProperty,errorMessagesHandler);
       thread = new Thread(dispacherThread);
       thread.start();
   }
   
 
   /**
    * Treat publications from subscribers
    * @param dataMessage
    */
   protected void processPublishedData(DataMessage dataMessage)
   {
   	// Repassar para a aplicao proxy a mensagem,
   	// e esta que passa para o servidor por exemplo... 
   	// ou simplesmente manda publicar pelo mtodo do proxy
   	
   	// TODO Por enquanto no tem EventProperties
   
   	ProxyAppNotify appNotify = new ProxyAppNotify(dataMessage);
      appNotify.start();
   }
		
	private class ProxyAppNotify extends Thread{

		DataMessage dataMessage;
		
		ProxyAppNotify (DataMessage dataMessage){
			this.dataMessage =  dataMessage;
		}
		
		public void run(){
			clientCommunic.notifyPublishedMessage(dataMessage.getSubject() ,dataMessage.getData(), protocol);
		}
	}
}
