package moca.core.proxy.receivers;

import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;
import moca.core.proxy.ClientInfo;
import moca.core.proxy.cache.Cacher;

/**
 * Este receiver  responsvel por analisar situaes de cache
 * Ele recebe mensagem especiais que tem por objetivo ativar e desativar
 * o estado de cache de um usurio.
 * @author hana
 */
public class CacheManager extends Receiver {
	/**
	 * Logger para o receiver
	 */
	private static final Logger LOGGER = Logger.getLogger(CacheManager.class);

	/**
	* Mapa que armazena os gerenciadores de cache para cada um dos usuarios
	*/
	private Map cacheMap = new HashMap();

	/**
	* Inicia o cache para um usuario
	* @param clientId identificador do usuario
	*/
	private synchronized void cacheStop(String clientId) {
		//recupera o mapa de cache e remov-o caso exista
		Cacher cacher = (Cacher) cacheMap.remove(clientId);
		if (cacher != null) {
			//se existe um objeto gerenciador de cache, recupere todas
			//as mensagem armazenadas e encaminhe-as para o objeto despachante
			Package[] packs = cacher.getPackages();
			disp.addPackages(packs);
		}
	}
	/**
	 * Inicia  cache para um cliente
	 * @param clientId identificador do cliente
	 * @param cacheClassName nome da classe de cache(Cacher) para armazenar as mensagens
	 */
	private synchronized void cacheStart(
		String clientId,
		String cacheClassName) {
		if (this.cacheMap.containsKey(clientId)) {
			//se o cliente j est no estado de cache, no faa nada
			return;
		}
		Cacher cacher;
		try {
			Class cacheClass = Class.forName(cacheClassName);
			cacher = (Cacher) cacheClass.newInstance();
		} catch (Exception e) {
			LOGGER.error(
				"Classe " + cacheClassName + " nao pode ser instanciada!",
				e);
			return;
		}
		//se tudo deu certo, adicione a classe armazenadora de mensagem no 
		//mapa de cache
		cacheMap.put(clientId, cacher);
	}
	/**
	 * Destroi as mensagens armazenadas do client 'clientId'
	 * @param clientId id do client
	 */
	private synchronized void cacheDestroy(String clientId) {
		//recupera o mapa de cache e remov-o caso exista
		cacheMap.remove(clientId);
	}
	
	/**
	 * Trata mensagens encaminhadas para avaliao pelo cache manager
	 * @param clientId identificador do usuario
	 * @param pack pacote
	 */
	private synchronized void treatPackage(String clientId, Package pack) {
		//recupera a classe de armazenamento das mensagens
		Cacher cacher = (Cacher) this.cacheMap.get(clientId);
		if (cacher == null) {
			//se ela no existir (no est sendo feito cache para o cliente), direcione a
			//mensagem para o adaptao
			pack.setDestination(AdapterManager.class);
			disp.addPackage(pack);
		} else {
			//caso esteja sendo feito cache para o cliente (marque o destino do usuario para o 
			//cache Manager novamente e armazene-o na classe cacher correspondente.  necessario
			//direcionar estas mensagens novamente para o cache manager porque quando da liberao destas
			//mensagens, muitas sero despachadas de uma unica vez, o que pode demorar algum tempo.
			//Neste intervalo, o cliente pode entrar novamente no estado de cache.
			pack.setDestination(CacheManager.class);
			cacher.addPackage(pack);
		}
	}
	/**
	 * Processa as mensagens enviadas ao cache manager. No momento, o cache manager esta preparado 
	 * para receber 3 tipos de mensagens.
	 * 
	 *  <ul>
	 *    <ui> CACHE_START   inicia o cache para um cliente
		 *    <ui> CACHE_STOP    finaliza o cache, retornando as mensagens armazenadas 
		 * 		   para o objeto despachante
	 *    <ui> CACHE_MESSAGE mensagem direcionada ao cache para avaliao. Se o cliente
	 * 		   destinatario da mensagem estiver em estado de cache, ela ser armazenada. 
	 * 		   Do contrario, ser direcionada para o proximo receiver
	 *  </ul>
	 */
	public final void process() {
		while (true) {
			//recupera o pacote
			Package pack = this.getNextPackage();
			//recupera as informaes do destinatario
			ClientInfo info = pack.getClientInfo();
			//recupera o identificador do usuario
			String clientId = info.getClientId();
			//recupera a mensagem
			ReceiverMessage msg = pack.getMessage();
			CacheMessage cm;
			if (msg instanceof CacheMessage) {
				//se a mensagem for uma mensagem de controle
				//do cache, faa o cast
				cm = (CacheMessage) msg;
			} else {
				//caso contrario, instancie uma mensagens de controle	
				cm = new CacheMessage(pack);
			}
			//avalie o tipo da mensagem de controle e direcione para
			//o mtodo tratador adequado.
			switch (cm.getType()) {
				//mensagem de cliente para avaliao
				case CacheMessage.CACHE_MESSAGE :
					this.treatPackage(clientId, pack);
					break;
				//mensagem de controle para iniciar o cache
				case CacheMessage.CACHE_START :
					this.cacheStart(clientId, cm.getCacheClassName());
					break;
				//mensagem de controle para finalizacao do cache
				case CacheMessage.CACHE_STOP :
					this.cacheStop(clientId);
					break;
           //	mensagem de controle para destruir o cache
				case CacheMessage.CACHE_DESTROY :
					this.cacheDestroy(clientId);
					break;
			}
		}
	}
	protected ReceiverResult process(ClientInfo info, ReceiverMessage msg) {
		//nao usado
		return null;
	}
}
