package moca.core.proxy.state;

import org.apache.log4j.Logger;

import moca.core.proxy.ClientInfo;
import moca.core.proxy.actions.Action;
import moca.core.proxy.actions.adapters.AdaptationException;
import moca.core.proxy.actions.adapters.Adapter;
import moca.core.proxy.actions.adapters.AggregatedAdapter;
import moca.core.proxy.message.Message;
import moca.core.proxy.state.filter.Filter;
import moca.core.proxy.state.filter.MetaPackage;

/**
 * Classe que representa uma regra. Uma regra determina um conjunto de aes a serem
 * aplicadas  uma mensagem, mediante critrios verificados sobre informaes da prpria
 * mensagem e/ou seu destinrio 
 * @author hana
 */
public class Rule implements Comparable {
	/**
	* Logger para Rule
	*/
	private static final Logger LOGGER = Logger.getLogger(Rule.class);
	/**
	* identificador da regra.
	*/
	private int id;
	/**
	* prioridade. Regras podem possuir prioridade, que permite estabelecer uma ordem
	* na aplicao das regras, independente da ordem de cadastro no arquivo de configurao.
	*/
	private int priority;
	/**
	 * Filtro. Contm uma expresso que avalia valores da mensagem e/ou do seu destinatario
	 * para verificar se a regra ser ou no aplicada
	 */
	private Filter f;
	/**
	* Adapter que agrega todos adapters cadastrados para esta regra
	*/
	private AggregatedAdapter agAdapter = new AggregatedAdapter();
	/**
	 * Adapter que verifica se o filtro permite ou no a aplicao desta regra sobre a
	 * mensagem 
	 */
	private Adapter adapter = new RuleAdapter();
	/**
	 * Determina uma ordem para dois inteiros. 
	 * 		Se x < y retorna -1
	 * 		Se x > y retorna 1
	 * 		Se x == y retorna 0
	 * @param x inteiro
	 * @param y inteiro
	 * @return inteiro representando a ordem entre os argumentos (-1, 0, 1)
	 */
	private int order(int x, int y) {
		return ((x < y) ? -1 : (x > y ? 1 : 0));
	}
	/**
	 * ordena as regras. Como a ordem define a prioridade, valores maiores
	 * devem vir primeiro na lista. Assim, a ordenacao sera no decrescente.
	 */
	public int compareTo(Object obj) {
		Rule otherRule = (Rule) obj;
		//ordenar por prioridade
		int order = this.order(this.priority, otherRule.priority);
		if (order != 0) {
			//ordenacao no decrescente
			return -order;
		}
		//ordenar por id (ordem de declaracao)
		return -order(this.id, otherRule.id);
	}

	public void setId(int id) {
		this.id = id;
	}
	public int getId() {
		return id;
	}

	private class RuleAdapter extends Adapter {
		public Message execute(ClientInfo clientInfo, Message msg)
			throws AdaptationException {
LOGGER.debug( "Cliente " + clientInfo.getClientId() +  " " + " Protocol: " + clientInfo.getCommunicationProtocol().getProtocol()+ " Data: " + msg.getDataType()); //%%%%%%
			if (f != null && !f.accepts(new MetaPackage(clientInfo, msg))) {
LOGGER.debug( " Filtro bloqueou a adaptao !!!! " ); //%%%%%%
				return msg;
			}
			return agAdapter.execute(clientInfo, msg);
		}
	}

	public Rule(Filter f, int priority) {
		this.f = f;
		this.priority = priority;
	}
	/**
	 * Adiciona um action a rule
	 * @param action classe que armazena informaes de configurao do action
	 */
	public void addAction(ActionConf action) {
		Action act = action.getAction();
		if (!(act instanceof Adapter)) {
			LOGGER.warn(
				"Classe "
					+ act
					+ " no  instancia de adapter! Ser descartada!");
		} else {
			this.agAdapter.addAdapter((Adapter) act);
		}
	}
	/**
	 * Retorna o adapter para a regra. O apater retornado  uma agregacao 
	 * de um ou mais adapters e s desempenha algum ao sobre a mensagem se o filtro
	 * aceita a mensagem
	 * @return adapter da regra 
	 */
	public Adapter getAdapter() {
		return this.adapter;
	}
	/**
	* Compara dois objetos
	*/
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (obj.getClass() != this.getClass()) {
			return false;
		}
		Rule rule = (Rule) obj;
		if (this.id == rule.id
			&& this.priority == rule.priority
			&& this.agAdapter.equals(rule.agAdapter)) {
			if ((this.f == null && rule.f == null)
				|| (this.f != null && this.f.equals(rule.f))) {
				return true;
			}
		}
		return false;
	}

}
