package moca.core.proxy.state;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;
import moca.core.proxy.ClientInfo;
import moca.core.proxy.actions.listeners.StateListener;
import moca.core.proxy.actions.listeners.StateListenerException;
import moca.core.proxy.receivers.Dispatcher;
import org.apache.log4j.Logger;

/**
 * Classe que representa um estado de monitoramento. O arquivo de configurao do proxy
 * possui um ou mais estados que sero monitorados pelo proxy, com o objetivo de adequar
 * as mensagens enviadas  situao do cliente.
 *  
 * @author hana
 *
 */
public class StateConf {
	/**
	 * StateConf logger
	 */
	private static final Logger LOGGER = Logger.getLogger(StateConf.class);
	/**
	* Conjunto de regras do estado.  necessario um conjunto ordenado.
	*/
	private Set rules = new TreeSet();
	/**
	 * Listeners cadastrados
	 */
	private List actions = new ArrayList();
	/**
	 * Expresso de monitoramento
	 */
	private Expression exp;
	/**
	 * Construtor. Um state sempre devera possuir uma expresso de monitoramento
	 * @param exp
	 */
	public StateConf(Expression exp) {
		this.exp = exp;
	}
	/**
	 * Adiciona uma regra. Uma regra define uma ou mais aes que sero aplicadas 
	 * uma mensagem de acordo com critrios (filtro) avaliados sobre a prpria mensagem ou 
	 * sobre seu destinatrio 
	 * @param rule regra a ser adicionada
	 */
	public void addRule(Rule rule) {
		//as regras recebem identificao na ordem em que so adicionadas
		int size = rules.size();
		rule.setId(size);
		this.rules.add(rule);
	}
	/**
	 * Retorna o conjunto de rules 
	 * @return rule set
	 */
	public Set getRuleSet() {
		return this.rules;
	}
	/**
	 * Retorna a expression associada ao estado
	 * @return expression
	 */
	public Expression getExpression() {
		return this.exp;
	}
	/**
	 * Adiciona um action ao estado. Os Actions adicionados sero sempre do tipo StateListener
	 * @param action ao para o estado. O estado apenas suporta aes do tipo StateListener
	 */
	public void addAction(ActionConf action) {
		this.actions.add(action);
	}
	/**
	 * Executa os listeners associados ao estado. Este mtodo  executado sempre
	 * que o estado do cliente muda de on para off ou de off para on
	 * @param isOn true se o estado se tornou on e false se o estado se tornou off
	 * @param disp objeto despachante
	 * @param info informaes sobre o cliente cujo estado se tornou on
	 */
	public void runListeners(boolean isOn, Dispatcher disp, ClientInfo info) {
		if (isOn) {
			for (int i = 0; i < actions.size(); i++) {
				ActionConf actionConf = (ActionConf) this.actions.get(i);
				StateListener listener = (StateListener) actionConf.getAction();
				try {
					listener.matches(info, disp);
				} catch (StateListenerException e) {
					//erro no listener
					LOGGER.warn(
						"Erro na execuao de "
							+ listener.getClass()
							+ ".matches()",
						e);
				}
			}
		} else {
			for (int i = 0; i < actions.size(); i++) {
				ActionConf actionConf = (ActionConf) this.actions.get(i);
				StateListener listener = (StateListener) actionConf.getAction();
				try {
					listener.unmatches(info, disp);
				} catch (StateListenerException e) {
					//erro no listener
					LOGGER.warn(
						"Erro na execuao de "
							+ listener.getClass()
							+ ".unmatches()",
						e);
				}
			}
		}
	}
	/**
	* Compara dois objetos
	*/
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (obj.getClass() != this.getClass()) {
			return false;
		}
		StateConf stConf = (StateConf) obj;
		if (stConf.actions.equals(this.actions)
			&& stConf.exp.equals(this.exp)
			&& this.rules.size() == stConf.rules.size()) {
			//necessario porque o equals de Set usa compare ao
			//inves de equals. Porm, neste contexto  necessario
			//verificar tambm os adapters.
			Iterator it1 = this.rules.iterator();
			Iterator it2 = stConf.rules.iterator();
			while (it1.hasNext()) {
				if (!it1.next().equals(it2.next())) {
					return false;
				}
			}
			return true;
		}
		return false;
	}
}
